/*
  Infiksavaldise ttleja
  Autor: Dan Bogdanov
  12.07.2000

  24.09.2002: parandasin vea unaarse miinuse lugemisel
  
  Vljakutse funktsioonile arvuta() teeb parameetrina antud avaldisele
  leksilise analsi, mille tulemused pannakse arvude ja tehtemrkide
  magasinidesse, kust siis igale tehtemrgile alates kige krgemast
  (sulge arvestades) vetakse kaks arvu ja tehakse tehted. Tulemus
  salvestatakse muutujasse tulemus. Funktsiooni enda vrtus on 'true'
  kui kik nnestus ja vigu ei olnud. Muidu 'false'.

  Vljakutse funktsioonile omista() tahab esimese parameetrina vikethte
  (muutuja nime) ja teisena avaldist, mille vrtus tuleb talle omistada.
  Samamoodi on funktsiooni vrtus edu puhul 'true'.
 
  Vimalikud lisandused:
  Luua baasklass avaldis, millest tuletada infiks-, prefiks-, ja
  postfiksavaldised. Mitmed universaalsed funktsioonid saab jtta
  baasklassi (virtuaalsetena) ja igale kujule ainuomased kirjutada
  vastavasse klassi.
*/

#include <stdio.h>
#include <avaldis.h>
#include <leksator.h>
#include <magasin.h>

//--Konstruktor---------------------------------------------------------------

infiksavaldis::infiksavaldis()
{
  analysaator=new leksator();
  am=new magasin();
  tm=new magasin();
  for(int i=0; i<=25; i++) {
    muutuja[i]=0;
    puudutatud[i]=false;
  }
}

//--Destruktor----------------------------------------------------------------

infiksavaldis::~infiksavaldis()
{
  delete analysaator;
  delete am;
  delete tm;
}

//--Funktsioon, mis arvutab avaldise vrtuse---------------------------------

int infiksavaldis::arvuta(char* uus_avaldis)
{
  // Valmistan kik arvutusteks ette
  analysaator->alusta(uus_avaldis);
  am->tyhjenda();
  tm->tyhjenda();
  ootanarvu=true;
  tulemus=0;

  // Panen magasini tehte algust thistava lisasulu, millega
  // vlistan segaduse tekkimise sulgude teineteise sisse jmisel
  lekseem* leks=new lekseem;
  leks->tyyp=L_MARK;
  leks->sisu=(char*)'(';
  tm->pane(leks);

  // Alustan tsklit, mis kestab kuni lekseemide (ja avaldise) lpuni vi
  // lpliku veani.
  do {
    // Loen lekseemi ja ttlen teda vastavalt tbile
    leks=analysaator->loelekseem();

    switch (leks->tyyp) {

      case L_ARV:
        // Kas praegu peaks jrjekorras olema arv?
        if(ootanarvu==false) {
          printf("Viga: Ootamatu arv!\n");
          return false;
        }
        else {
        // Kas tegemist on negatiivse arvuga?
          // Panen arvu arvude magasini
          am->pane(leks);
          ootanarvu=false;
        }
        break;

      case L_MUUT:
        // Kas praegu vib tulla muutuja?
        if(ootanarvu==false) {
          printf("Viga: Ootamatu muutuja!\n");
          return false;
        }
        else {
          // Kas muutujale on vrtus omistatud?
          // Kui jah, siis muuda lekseem arvuks
          if(puudutatud[(char)leks->sisu-'a']) {
            leks->tyyp=L_ARV;
            leks->arv=muutuja[(char)leks->sisu-'a'];
          }
          else {
            printf("Viga: Muutuja vrtustamata!\n");
            return false;
          }
          // Kas muutuja ees on miinusmrk?
          // Pane saadud arv arvude magasini
          am->pane(leks);
          ootanarvu=false;
        }
        break;

      case L_MARK:
        // Kitun vastavalt mrgile
        switch ((char)leks->sisu) {

          case '(':
            if(ootanarvu==false) {
              printf("Viga: '(' tehtemrgi asemel!\n");
              return false;
            }
            else
              tm->pane(leks);
            break;

          case ')':
            if(ootanarvu==true) {
              printf("Viga: ')' tehtemrgi asemel!\n");
              return false;
            }
            else {
              // Tegeleme sulu sisuga
              if(tootlemark(leks)==false)
                return false;
            }
            break;

          case '+':
          case '-':
          case '*':
          case '/':
          // Kontrollin, kas arvu asemel on tehtemrk (mis pole miinus)
          // Kui on miinusmrk, thendab jrgnev arv peab tulema negatiivne
          if(ootanarvu==true && (char)leks->sisu!='-') {
            printf("Viga: Tehtemrk arvu asemel!\n");
            return false;
          } else {
            if(ootanarvu==true && (char)leks->sisu=='-') {
              lekseem* null = new lekseem ();
              null->tyyp=L_ARV;
              null->sisu=NULL;
              null->arv=0;
              am->pane (null);
            }

            if(tootlemark(leks)==false)
              return false;
            ootanarvu=true;
          }
        }
        break;
      case L_VIGA:
        printf("Viga: %s\n", leks->sisu);
        return false;
        break;
    }
  } while(leks->tyyp!=L_LOPP);

  // Vaatan, kas peale avaldise lppu ei tunne me puudust veel mne arvu
  // jrele
  if(ootanarvu==true) {
    printf("Viga: Avaldise lpp arvu asemel!\n");
    return false;
  }
  else {
    // Lisan avaldise lppu sulgeva sulu ja saadab kogu komplekti tsse
    leks->tyyp=L_MARK;
    leks->sisu=(char*)')';
    if(tootlemark(leks)==false)
      return false;
  }

  if(leks->tyyp!=L_VIGA) {
    // Vta magasinist viimane arv - s.o. avaldise vrtus
    leks=am->vota();
    if(leks->tyyp==L_ARV) {
      if(!am->on_tyhi() || !tm->on_tyhi()) {
        printf("Viga: Magasin pole thi! (kontrolli sulge)\n");
        return false;
      }
    }
  }

  // Kui avaldise vrtuse asemel oli magasinis midagi muud, peab see
  // olema viga
  if(leks->tyyp!=L_ARV) {
    printf("Viga: %s\n",leks->sisu);
    return false;
  }
  tulemus=leks->arv;
  return true;
}

//--Omistab muutujale vrtuse------------------------------------------------

int infiksavaldis::omista(char taht, char* uus_avaldis)
{
  int edu;
  // Kas muutuja sobib meie tingimustega?
  if(taht>='A' && taht <='Z' || taht>='a' && taht <='z') {
    // Arvutame avaldise vrtuse
    edu=arvuta(uus_avaldis);
    // Kui see on teada, seame kik korda muutujate tabelis
    if(edu==true) {
      muutuja[taht-'a']=tulemus;
      puudutatud[taht-'a']=true;
    }
  }
  else {
  printf("Viga: Lubamatu muutujanimi!\n");
  edu=false;
  }
  return edu;
}

//--Mrab vastavalt tehtele tema prioriteedi---------------------------------

int infiksavaldis::prioriteet(lekseem *leks)
{
  // Kige krgemal on tehete puus korrutamine/jagamine. Need tuleb teha
  // viimasena
  switch ((char)leks->sisu) {
    case '(': return 0;
    case ')': return 1;
    case '+': case '-': return 2;
    case '*': case '/': return 3;
  }
  return -1;
}

//--Kui magasin on kokku pandud, vtab sealt arve ja sooritab tehted----------

int infiksavaldis::soorita(lekseem* leks)
{
  // Vtan magasinist kaks arvu
  lekseem* l2=am->vota();
  lekseem* l1=am->vota();
  if(l1->tyyp==L_ARV && l2->tyyp==L_ARV) {
    // Tehtemrki hakkab asendama arv, mis antud tehte tulemuseks on
    leks->tyyp=L_ARV;
    switch ((char)leks->sisu) {
      // Teeme tehted
      case '+': leks->arv=l1->arv + l2->arv; break;
      case '-': leks->arv=l1->arv - l2->arv; break;
      case '*': leks->arv=l1->arv * l2->arv; break;
      case '/':
        // Kontrollime jagamist nulliga
        if(l2->arv==0) {
          printf("Viga: Jagamine nulliga!\n");
          return false;
        }
        else {
          leks->tyyp=L_ARV;
          leks->arv=l1->arv / l2->arv;
        }
        break;
    }
    // See tulemus lheb juba arvude magasini
    am->pane(leks);
    return true;
  }
  else {
    printf("Viga: Magasin thi!\n");
  }
  return false;
}

//--Vastavalt tehtemrkidele sooritab tehteid ja haldab jrjekorda------------

int infiksavaldis::tootlemark(lekseem *leks)
{
  lekseem* ll;

  do {
    // Vtame kaks tehtemrki ja nende suhte jrgi vaatame, kas ja kui
    // siis millisega neist peaks tehet sooritama
    ll=tm->vota();
    if(ll->tyyp==L_MARK && prioriteet(leks)<=prioriteet(ll)) {
      if(soorita(ll)==false)
        return false;
    }
  } while(((ll->tyyp!=L_VIGA) && (ll->tyyp!=L_MARK)) || prioriteet(leks)<=prioriteet(ll));
  // Sooritatud tehte mrk lheb kaduma
  if(ll->tyyp==L_MARK && ((char)leks->sisu)!=')') {
    tm->pane(ll);
    tm->pane(leks);
  }
  return true;
}

//--Lpp----------------------------------------------------------------------

